# Copyright 2026 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2021 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# mypy: disable-error-code=method-assign

from __future__ import annotations

import os
import pathlib
import re
import sys
import textwrap
import uuid
from unittest.mock import MagicMock, patch

import cirq
import cirq_superstaq as css
import numpy as np
import pandas as pd
import pytest

from supermarq.qcvv.su2 import SU2, SU2Results, decay


@pytest.fixture(scope="session", autouse=True)
def patch_tqdm() -> None:
    os.environ["TQDM_DISABLE"] = "1"


def test_su2_init() -> None:
    with patch("cirq_superstaq.service.Service"):
        experiment = SU2(4, [1, 2, 3, 4], cirq.CNOT)
    assert experiment.num_qubits == 2
    assert experiment.two_qubit_gate == cirq.CNOT
    assert experiment.num_circuits == 4
    assert experiment.cycle_depths == [1, 2, 3, 4]


def test_su2_init_raises() -> None:
    with (
        patch("cirq_superstaq.service.Service"),
        pytest.raises(
            ValueError,
            match=r"The `two_qubit_gate` parameter must be a gate that acts on exactly two qubits.",
        ),
    ):
        SU2(4, [1, 2, 3, 4], cirq.X)


@pytest.fixture
def su2_experiment() -> SU2:
    with patch("cirq_superstaq.service.Service"):
        return SU2(4, [1, 2, 3, 4], cirq.CNOT)


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Python version < 3.10")
def test_build_circuits(su2_experiment: SU2) -> None:  # pragma: no cover
    # By extension also tests the _component() method
    # Patch the PhasedXZ drawing function to ignore the rotation parameters
    cirq.PhasedXZGate._circuit_diagram_info_ = MagicMock(
        return_value=cirq.CircuitDiagramInfo(wire_symbols=["PhXZ"])
    )

    samples = su2_experiment._build_circuits(2, [1, 3])
    assert len(samples) == 4
    for sample in samples:
        assert "num_two_qubit_gates" in sample.data
        assert sample.data["num_two_qubit_gates"] in [2, 6]
        if sample.data["num_two_qubit_gates"] == 2:
            cirq.testing.assert_has_diagram(
                sample.circuit,
                textwrap.dedent(
                    """
                    0: ───PhXZ───│───@[no_compile]───│───X───│───@[no_compile]───│───PhXZ───│───X───│───PhXZ───│───X───│───PhXZ───M───
                                 │   │               │       │   │               │          │       │          │       │          │
                    1: ───PhXZ───│───X───────────────│───X───│───X───────────────│───PhXZ───│───X───│───PhXZ───│───X───│───PhXZ───M───
                    """  # noqa: E501
                ),
            )
        else:
            cirq.testing.assert_has_diagram(
                sample.circuit,
                textwrap.dedent(
                    """
                    0: ───PhXZ───│───@[no_compile]───│───X───│───@[no_compile]───│───PhXZ───│───@[no_compile]───│───X───│───@[no_compile]───│───PhXZ───│───@[no_compile]───│───X───│───@[no_compile]───│───PhXZ───M───
                                 │   │               │       │   │               │          │   │               │       │   │               │          │   │               │       │   │               │          │
                    1: ───PhXZ───│───X───────────────│───X───│───X───────────────│───PhXZ───│───X───────────────│───X───│───X───────────────│───PhXZ───│───X───────────────│───X───│───X───────────────│───PhXZ───M───
                    """  # noqa: E501
                ),
            )


@pytest.mark.skipif(sys.version_info >= (3, 10), reason="Python version >= 3.10")
def test_build_circuits_old(su2_experiment: SU2) -> None:  # pragma: no cover
    # By extension also tests the _component() method
    # Patch the PhasedXZ drawing function to ignore the rotation parameters
    cirq.PhasedXZGate._circuit_diagram_info_ = MagicMock(
        return_value=cirq.CircuitDiagramInfo(wire_symbols=["PhXZ"])
    )

    samples = su2_experiment._build_circuits(2, [1, 3])
    assert len(samples) == 4
    for sample in samples:
        assert "num_two_qubit_gates" in sample.data
        assert sample.data["num_two_qubit_gates"] in [2, 6]
        if sample.data["num_two_qubit_gates"] == 2:
            cirq.testing.assert_has_diagram(
                sample.circuit,
                textwrap.dedent(
                    """
                    0: ───PhXZ───│───@['no_compile']───│───X───│───@['no_compile']───│───PhXZ───│───X───│───PhXZ───│───X───│───PhXZ───M───
                                 │   │                 │       │   │                 │          │       │          │       │          │
                    1: ───PhXZ───│───X─────────────────│───X───│───X─────────────────│───PhXZ───│───X───│───PhXZ───│───X───│───PhXZ───M───
                    """  # noqa: E501
                ),
            )
        else:
            cirq.testing.assert_has_diagram(
                sample.circuit,
                textwrap.dedent(
                    """
                    0: ───PhXZ───│───@['no_compile']───│───X───│───@['no_compile']───│───PhXZ───│───@['no_compile']───│───X───│───@['no_compile']───│───PhXZ───│───@['no_compile']───│───X───│───@['no_compile']───│───PhXZ───M───
                                 │   │                 │       │   │                 │          │   │                 │       │   │                 │          │   │                 │       │   │                 │          │
                    1: ───PhXZ───│───X─────────────────│───X───│───X─────────────────│───PhXZ───│───X─────────────────│───X───│───X─────────────────│───PhXZ───│───X─────────────────│───X───│───X─────────────────│───PhXZ───M───
                    """  # noqa: E501
                ),
            )


def test_analyse_results(tmp_path: pathlib.Path, su2_experiment: SU2) -> None:
    result = SU2Results(
        target="example",
        experiment=su2_experiment,
        data=pd.DataFrame(
            [
                {
                    "num_two_qubit_gates": 2,
                    "circuit_realization": 1,
                    "uuid": uuid.uuid4(),
                    "00": decay(2, 0.25, 0.975),
                    "01": 0.0,
                    "10": 0.0,
                    "11": 1 - decay(2, 0.25, 0.975),
                },
                {
                    "num_two_qubit_gates": 4,
                    "circuit_realization": 1,
                    "uuid": uuid.uuid4(),
                    "00": decay(4, 0.25, 0.975),
                    "01": 0.0,
                    "10": 0.0,
                    "11": 1 - decay(4, 0.25, 0.975),
                },
                {
                    "num_two_qubit_gates": 6,
                    "circuit_realization": 1,
                    "uuid": uuid.uuid4(),
                    "00": decay(6, 0.25, 0.975),
                    "01": 0.0,
                    "10": 0.0,
                    "11": 1 - decay(6, 0.25, 0.975),
                },
                {
                    "num_two_qubit_gates": 8,
                    "circuit_realization": 1,
                    "uuid": uuid.uuid4(),
                    "00": decay(8, 0.25, 0.975),
                    "01": 0.0,
                    "10": 0.0,
                    "11": 1 - decay(8, 0.25, 0.975),
                },
            ]
        ),
    )

    plot_filename = tmp_path / "example.png"

    result.analyze(plot_filename=plot_filename.as_posix())

    assert pathlib.Path(tmp_path / "example.png").exists()

    assert result.two_qubit_gate_fidelity == pytest.approx(0.975, abs=1e-6)
    assert result.two_qubit_gate_fidelity_std == pytest.approx(0.0, abs=1e-6)
    assert result.two_qubit_gate_error == pytest.approx(0.025, abs=1e-6)
    assert result.two_qubit_gate_error_std == pytest.approx(0.0, abs=1e-6)
    assert result.single_qubit_noise == pytest.approx(0.25, abs=1e-6)
    assert result.single_qubit_noise_std == pytest.approx(0.0, abs=1e-6)


def test_haar_random_rotation() -> None:
    r = SU2._haar_random_rotation()
    assert isinstance(r, cirq.Gate)
    assert r.num_qubits() == 1

    with patch("cirq.testing.random_special_unitary", return_value=np.eye(2)):
        assert SU2._haar_random_rotation() == cirq.I


def test_result_not_analyzed() -> None:
    result = SU2Results(target="example", experiment=MagicMock(spec=SU2))

    for attr in [
        "two_qubit_gate_fidelity",
        "two_qubit_gate_fidelity_std",
        "two_qubit_gate_error",
        "two_qubit_gate_error_std",
        "single_qubit_noise",
        "single_qubit_noise_std",
    ]:
        with pytest.raises(
            RuntimeError,
            match=re.escape("Value has not yet been estimated. Please run `.analyze()` method."),
        ):
            _ = getattr(result, attr)


def test_result_missing_data() -> None:
    result = SU2Results(target="example", experiment=MagicMock(spec=SU2))

    with pytest.raises(RuntimeError, match=r"No data stored. Cannot perform analysis."):
        result._analyze()

    with pytest.raises(RuntimeError, match=r"No data stored. Cannot plot results."):
        result.plot_results()


def test_dump_and_load(
    tmp_path_factory: pytest.TempPathFactory,
    su2_experiment: SU2,
) -> None:
    filename = tmp_path_factory.mktemp("tempdir") / "file.json"
    su2_experiment.to_file(filename)
    exp = SU2.from_file(filename)

    assert exp.samples == su2_experiment.samples
    assert exp.num_qubits == su2_experiment.num_qubits
    assert exp.num_circuits == su2_experiment.num_circuits
    assert exp.cycle_depths == su2_experiment.cycle_depths
    assert exp.two_qubit_gate == su2_experiment.two_qubit_gate


def test_repr(su2_experiment: SU2) -> None:
    assert su2_experiment.__repr__() == "SU2(two_qubit_gate=CNOT, num_qubits=2, num_samples=16)"
    results = SU2Results(target="target", experiment=su2_experiment, job=MagicMock(spec=css.Job))
    assert results.__repr__() == (
        "SU2Results(Results not analyzed, experiment=SU2(two_qubit_gate=CNOT, num_qubits=2, "
        "num_samples=16), target=target)"
    )

    results._two_qubit_gate_fidelity = 0.98
    results._two_qubit_gate_fidelity_std = 0.01
    results._single_qubit_noise = 0.999
    results._single_qubit_noise_std = 0.1
    assert results.__repr__() == (
        "SU2Results(Estimated two qubit gate fidelity: 0.98 +/- 0.01, Estimated "
        "single qubit noise: 0.999 +/- 0.1, experiment=SU2(two_qubit_gate=CNOT, num_qubits=2, "
        "num_samples=16), target=target)"
    )



# Copyright 2026 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .qiskit import cirq_gate_to_qiskit_gate, cirq_to_qiskit, qiskit_to_cirq

__all__ = [
    "cirq_gate_to_qiskit_gate",
    "cirq_to_qiskit",
    "qiskit_to_cirq",
]



# Copyright 2026 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tooling for Symmetric Stabilizer Benchmarking.
See https://arxiv.org/pdf/2407.20184 for more details.
"""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import cirq
import cirq_superstaq as css
import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize
import seaborn as sns
import tqdm
import tqdm.contrib
import tqdm.contrib.itertools

from supermarq.qcvv.base_experiment import QCVVExperiment, QCVVResults, Sample

if TYPE_CHECKING:
    import numpy.typing as npt
    from typing_extensions import Self


def _exp_decay(
    x: npt.NDArray[np.float64], A: float, alpha: float, B: float
) -> npt.NDArray[np.float64]:
    r"""Exponential decay of the form.

    .. math::

        A \alpha^x + B

    Args:
        x: x
        A: Decay constant
        alpha: Decay coefficient
        B: Additive constant
    Returns:
        Exponential decay applied to x.
    """
    return A * (alpha**x) + B


@dataclass(repr=False)
class SSBResults(QCVVResults):
    """Results from an SSB experiment."""

    _cz_fidelity_estimate: float | None = None
    """Estimated CZ fidelity."""
    _cz_fidelity_estimate_std: float | None = None
    """Standard deviation for the CZ fidelity estimate."""
    _fit: tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]] | None = None
    """The fitted values"""

    @property
    def cz_fidelity_estimate(self) -> float:
        """Estimated CZ fidelity."""
        if self._cz_fidelity_estimate is None:
            raise self._not_analyzed
        return self._cz_fidelity_estimate

    @property
    def cz_fidelity_estimate_std(self) -> float:
        """Standard deviation for the CZ fidelity estimate."""
        if self._cz_fidelity_estimate_std is None:
            raise self._not_analyzed
        return self._cz_fidelity_estimate_std

    def _analyze(self) -> None:
        """Analyse the results and calculate the estimated CZ gate fidelity.

        Args:
            plot_results (optional): Whether to generate the data plots. Defaults to True.

        Returns:
           The final results from the experiment.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot perform analysis.")

        xx = self.data["num_cz_gates"]
        yy = self.data["0" * self.num_qubits]

        self._fit = scipy.optimize.curve_fit(
            _exp_decay,
            xx,
            yy,
            p0=(1.0 - 2**-self.num_qubits, 0.99, 2**-self.num_qubits),
            bounds=(0, 1),
            max_nfev=2000,
        )
        assert self._fit is not None

        self._cz_fidelity_estimate = self._fit[0][1]
        self._cz_fidelity_estimate_std = np.sqrt(self._fit[1][1][1])

    def plot_results(self, filename: str | None = None) -> plt.Figure:
        """Plot the experiment data and the corresponding fits. The shaded upper and lower limits
        of the shaded region indicate the fits at +/- 1 standard deviation in all fitted parameters.

        Args:
            filename: Optional argument providing a filename to save the plots to. Defaults to None,
                indicating not to save the plot.

        Returns:
            A single matplotlib figure with the experimental data and corresponding fits.

        Raises:
            RuntimeError: If there is no data stored.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot plot results.")
        assert self._fit is not None

        fig, axs = plt.subplots(1, 1)
        sns.scatterplot(
            data=self.data, x="num_cz_gates", y="0" * self.num_qubits, ax=axs, label="Data"
        )
        xx = np.linspace(0, np.max(self.data.num_cz_gates))
        axs.plot(xx, _exp_decay(xx, *self._fit[0]), color="tab:blue", linestyle="--", label="Fit")
        axs.fill_between(
            xx,
            _exp_decay(xx, *(self._fit[0] - np.sqrt(np.diag(self._fit[1])))),
            _exp_decay(xx, *(self._fit[0] + np.sqrt(np.diag(self._fit[1])))),
            alpha=0.35,
            color="tab:blue",
            label="1s.d C.I.",
        )
        axs.legend()

        axs.set_xlabel("Number of CZ Gates")
        axs.set_ylabel("Final state probability")

        if filename is not None:
            fig.savefig(filename, bbox_inches="tight")

        return fig

    def _results_msg(self) -> str:
        """Prints the key results data."""
        return (
            f"Estimated CZ fidelity: {self.cz_fidelity_estimate:.6} "
            f"+/- {self.cz_fidelity_estimate_std:.6}"
        )


class SSB(QCVVExperiment[SSBResults]):
    """Symmetric Stabilizer Benchmarking. A benchmarking algorithm for determining the CZ fidelity
    of a device. Specifically designed for neutral atom devices where CZ-gates mediated by Rydberg
    interactions are the native entangling gate.

    See: https://arxiv.org/abs/2407.20184
    """

    def __init__(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
        *,
        random_seed: int | np.random.Generator | None = None,
        _samples: list[Sample] | None = None,
        **kwargs: str,
    ) -> None:
        """Initializes a symmetric stabilizer benchmarking experiment.

        Args:
            num_circuits: Number of circuits to sample.
            cycle_depths: The cycle depths to sample.
            random_seed: An optional seed to use for randomization.
            kwargs: Additional kwargs passed to the base QCVVExperiment.
        """
        qubits = cirq.LineQubit.range(2)

        # Moments containing parallel rotations.
        X = css.ParallelRGate(np.pi / 2, 0.0, 2)
        Y = css.ParallelRGate(np.pi / 2, np.pi / 2, 2)
        _X = css.ParallelRGate(np.pi / 2, np.pi, 2)
        _Y = css.ParallelRGate(np.pi / 2, -np.pi / 2, 2)
        self._single_qubit_gate_set = [X, _X, Y, _Y]

        # Table I of https://arxiv.org/pdf/2407.20184
        stabilizer_states = np.array(
            [
                [1, 1, 1, 1],
                [1, -1, -1, 1],
                [1, 1j, 1j, -1],
                [1, -1j, -1j, -1],
                [1, 0, 0, 0],
                [0, 0, 0, 1],
                [1, 1, 1, -1],
                [1, -1, -1, -1],
                [1, 1j, 1j, 1],
                [1, -1j, -1j, 1],
                [1, 0, 0, 1j],
                [1, 0, 0, -1j],
            ]
        )
        stabilizer_states /= np.linalg.norm(stabilizer_states, axis=1, keepdims=True)
        self._stabilizer_states = stabilizer_states

        # Table II of https://arxiv.org/pdf/2407.20184
        self._init_rotations = [
            [_X, X, Y, _Y, Y],
            [_X, _X, Y, _Y, Y],
            [_X, _X, X, _X, X],
            [_X, X, X, _X, X],
            [_X, X, X, _Y, _X],
            [_X, _X, X, _Y, _X],
            [X, _Y, X, _Y, _X],
            [_X, Y, X, _Y, _X],
            [X, _Y, X, _X, _X],
            [_X, Y, X, _X, X],
            [_X, Y, Y, _Y, Y],
            [X, _Y, Y, _Y, Y],
        ]
        # Table III of https://arxiv.org/pdf/2407.20184
        self._reconciliation_rotation = [
            [X, _Y, X, _X],
            [X, Y, X, _X],
            [Y, X, X, _X],
            [Y, _X, X, _X],
            [_X, X, X, _X],
            [X, X, X, _X],
            [_X, Y, Y, _X],
            [X, Y, Y, _X],
            [Y, Y, Y, _X],
            [X, X, Y, _X],
            [_Y, X, Y, _X],
            [Y, X, Y, _X],
        ]

        super().__init__(
            qubits=qubits,
            num_circuits=num_circuits,
            cycle_depths=cycle_depths,
            random_seed=random_seed,
            results_cls=SSBResults,
            _samples=_samples,
            **kwargs,
        )

    ###################
    # Private Methods #
    ###################
    def _build_circuits(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
    ) -> Sequence[Sample]:
        """Build a list of random circuits to perform the SSB experiment with.

        Args:
            num_circuits: Number of circuits to generate.
            cycle_depths: An iterable of the different numbers of cycles to include in each circuit.

        Returns:
            The list of experiment samples.
        """
        random_circuits = []
        min_depth = min(cycle_depths)
        max_depth = max(cycle_depths)
        if min_depth < 2:
            raise ValueError("Cannot perform SSB with a cycle depth of 1.")

        # Precompute action of each gate on the stabilizer state index. This is just Clifford
        # tableau math but we only have to do this once though so there really isn't much of a
        # performance difference
        idx_maps = {}
        for gate in (*self._single_qubit_gate_set, cirq.CZ):
            mat = cirq.unitary(gate)
            inner_products = abs(self._stabilizer_states.conj() @ mat @ self._stabilizer_states.T)
            idx_maps[gate] = dict(enumerate(inner_products.argmax(0)))

        # Build samples
        for k, depth in tqdm.contrib.itertools.product(
            range(num_circuits), cycle_depths, desc="Building circuits"
        ):
            sss_idx = int(self._rng.integers(0, 11, endpoint=True))
            circuit = self._sss_init_circuit(sss_idx)

            circuit += css.Barrier(2)(*self.qubits)

            for i in range(max_depth - 2):
                gate = self._random_parallel_qubit_rotation()
                circuit += gate.on(*self.qubits)
                sss_idx = idx_maps[gate][sss_idx]

                if i < depth - 2:
                    circuit += cirq.CZ.on(*self.qubits).with_tags("no_compile")
                    sss_idx = idx_maps[cirq.CZ][sss_idx]

                circuit += css.Barrier(2)(*self.qubits)

            circuit += self._sss_reconciliation_circuit(sss_idx) + cirq.measure(self.qubits)

            random_circuits.append(
                Sample(
                    circuit=circuit,
                    data={
                        "num_cz_gates": depth,
                        "initial_sss_index": sss_idx,
                    },
                    circuit_realization=k,
                )
            )
        return random_circuits

    def _random_parallel_qubit_rotation(self) -> cirq.Gate:
        """Chooses randomly from {X, Y, -X, -Y} and return a moment with this gate acting on both
        quits. Note we don't use `cirq.ParallelGate` as when modelling noise Cirq treats this as
        a two qubit gate.

        Returns:
            The randomly chosen rotation gate acting on both qubits.
        """
        return self._rng.choice(np.array(self._single_qubit_gate_set))

    def _sss_init_circuit(self, idx: int) -> cirq.Circuit:
        """Creates the initialisation circuit for the provided symmetric-stabiliser state index.
        See appendix of https://arxiv.org/pdf/2407.20184 for details.

        Args:
            idx: The index of the desired symmetric-stabiliser state

        Returns:
            A circuit that maps the |00> state to the desired symmetric-stabiliser state.
        """
        init_circuit = cirq.Circuit(
            self._init_rotations[idx][0].on(*self.qubits),
            self._init_rotations[idx][1].on(*self.qubits),
            cirq.CZ(*self.qubits).with_tags("no_compile"),
            self._init_rotations[idx][2].on(*self.qubits),
            self._init_rotations[idx][3].on(*self.qubits),
            self._init_rotations[idx][4].on(*self.qubits),
        )
        return init_circuit

    def _sss_reconciliation_circuit(self, idx: int) -> cirq.Circuit:
        """Given a randomly generated circuit, return the appropriate reconciliation circuit.
        See appendix of https://arxiv.org/pdf/2407.20184 for details.

        Args:
            idx: The index of the final symmetric-stabiliser state

        Returns:
            The appropriate reconciliation circuit
        """
        # Return the reconciliation circuit. See table III of https://arxiv.org/pdf/2407.20184
        return cirq.Circuit(
            self._reconciliation_rotation[idx][0].on(*self.qubits),
            self._reconciliation_rotation[idx][1].on(*self.qubits),
            cirq.CZ(*self.qubits).with_tags("no_compile"),
            self._reconciliation_rotation[idx][2].on(*self.qubits),
            self._reconciliation_rotation[idx][3].on(*self.qubits),
        )

    def _json_dict_(self) -> dict[str, Any]:
        """Converts the experiment to a json-able dictionary that can be used to recreate the
        experiment object. Note that the state of the random number generator is not stored.

        Returns:
            Json-able dictionary of the experiment data.
        """
        return super()._json_dict_()

    @classmethod
    def _from_json_dict_(
        cls,
        _samples: list[Sample],
        num_circuits: int,
        cycle_depths: list[int],
        **kwargs: Any,
    ) -> Self:
        """Creates a experiment from a dictionary of the data.

        Args:
            dictionary: Dict containing the experiment data.

        Returns:
            The deserialized experiment object.
        """
        kwargs.pop("qubits")  # Don't need for SSB

        return cls(
            num_circuits=num_circuits,
            cycle_depths=cycle_depths,
            _samples=_samples,
            **kwargs,
        )



# Copyright 2026 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2021 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# mypy: disable-error-code=method-assign

from __future__ import annotations

import pathlib
import re
from unittest.mock import MagicMock

import cirq
import cirq_superstaq as css
import numpy as np
import pandas as pd
import pytest

from supermarq.qcvv import SSB, SSBResults


def test_ssb_init() -> None:
    q0, q1 = cirq.LineQubit.range(2)

    experiment = SSB(num_circuits=10, cycle_depths=[2, 3, 5])
    assert experiment.num_qubits == 2
    assert experiment.qubits == (q0, q1)

    assert len(experiment._stabilizer_states) == 12
    assert len(experiment._init_rotations) == 12
    for rot in experiment._init_rotations:
        assert len(rot) == 5
    assert len(experiment._reconciliation_rotation) == 12
    for rot in experiment._reconciliation_rotation:
        assert len(rot) == 4

    with pytest.raises(ValueError, match=r"Cannot perform SSB with a cycle depth of 1."):
        SSB(num_circuits=10, cycle_depths=[1, 2, 3, 5])


@pytest.fixture
def ssb_experiment() -> SSB:
    return SSB(num_circuits=10, cycle_depths=[2, 3, 5])


def test_random_parallel_qubit_rotation(ssb_experiment: SSB) -> None:
    ssb_experiment._rng = (rng := MagicMock())
    rng.choice.side_effect = [
        cirq.rx(np.pi / 2),
        cirq.ry(-np.pi / 2),
    ]
    moment = ssb_experiment._random_parallel_qubit_rotation()
    assert moment == cirq.rx(np.pi / 2)

    moment = ssb_experiment._random_parallel_qubit_rotation()
    assert moment == cirq.ry(-np.pi / 2)


def test_sss_init_circuits(ssb_experiment: SSB) -> None:
    init_circuit = ssb_experiment._sss_init_circuit(4)
    # Index-4 has init_rotations [X, X, X, _Y, _X]
    q0, q1 = cirq.LineQubit.range(2)
    assert init_circuit == cirq.Circuit(
        css.ParallelRGate(np.pi / 2, np.pi, 2)(q0, q1),
        css.ParallelRGate(np.pi / 2, 0, 2)(q0, q1),
        cirq.CZ(q0, q1).with_tags("no_compile"),
        css.ParallelRGate(np.pi / 2, 0, 2)(q0, q1),
        css.ParallelRGate(np.pi / 2, -np.pi / 2, 2)(q0, q1),
        css.ParallelRGate(np.pi / 2, np.pi, 2)(q0, q1),
    )


def test_sss_reconciliation_circuit(ssb_experiment: SSB) -> None:
    recon_circuit = ssb_experiment._sss_reconciliation_circuit(4)
    q0, q1 = cirq.LineQubit.range(2)
    assert recon_circuit == cirq.Circuit(
        css.ParallelRGate(np.pi / 2, np.pi, 2)(q0, q1),
        css.ParallelRGate(np.pi / 2, 0.0, 2)(q0, q1),
        cirq.CZ(q0, q1).with_tags("no_compile"),
        css.ParallelRGate(np.pi / 2, 0.0, 2)(q0, q1),
        css.ParallelRGate(np.pi / 2, np.pi, 2)(q0, q1),
    )


def test_build_ssb_circuit(ssb_experiment: SSB) -> None:
    X = css.ParallelRGate(np.pi / 2, 0.0, 2)
    _X = css.ParallelRGate(np.pi / 2, np.pi, 2)
    Y = css.ParallelRGate(np.pi / 2, np.pi / 2, 2)
    _Y = css.ParallelRGate(np.pi / 2, -np.pi / 2, 2)
    ssb_experiment._rng = (rng := MagicMock())
    rng.integers.return_value = 4
    rng.choice.side_effect = [_X, _Y]
    circuits = ssb_experiment._build_circuits(num_circuits=1, cycle_depths=[4])

    assert len(circuits) == 1

    q0, q1 = ssb_experiment.qubits
    # Init circuit
    expected_circuit = [
        cirq.Moment(_X(q0, q1)),
        cirq.Moment(X(q0, q1)),
        cirq.CZ(q0, q1).with_tags("no_compile"),
        cirq.Moment(X(q0, q1)),
        cirq.Moment(_Y(q0, q1)),
        cirq.Moment(_X(q0, q1)),
        css.Barrier(2)(q0, q1),
    ]
    # Intermediate ops
    expected_circuit += [
        cirq.Moment(_X(q0, q1)),
        cirq.CZ(q0, q1).with_tags("no_compile"),
        css.Barrier(2)(q0, q1),
        cirq.Moment(_Y(q0, q1)),
        cirq.CZ(q0, q1).with_tags("no_compile"),
        css.Barrier(2)(q0, q1),
    ]
    # Reconcilliation
    expected_circuit += [
        cirq.Moment(_Y(q0, q1)),
        cirq.Moment(X(q0, q1)),
        cirq.CZ(q0, q1).with_tags("no_compile"),
        cirq.Moment(Y(q0, q1)),
        cirq.Moment(_X(q0, q1)),
        cirq.Moment(
            cirq.measure(cirq.LineQubit(0), cirq.LineQubit(1)),
        ),
    ]

    cirq.testing.assert_same_circuits(
        circuits[0].circuit,
        cirq.Circuit(expected_circuit),
    )
    assert circuits[0].data == {
        "initial_sss_index": 10,
        "num_cz_gates": 4,
    }


def test_ssb_analyse_results(tmp_path: pathlib.Path, ssb_experiment: SSB) -> None:
    results = SSBResults(target="example", experiment=ssb_experiment)

    results.data = pd.DataFrame(
        [
            {
                "initial_sss_index": 4,
                "num_cz_gates": 2,
                "00": 1.0,
                "01": 0.0,
                "10": 0.0,
                "11": 0.0,
            },
            {
                "initial_sss_index": 4,
                "num_cz_gates": 3,
                "00": 0.9,
                "01": 0.0,
                "10": 0.0,
                "11": 0.0,
            },
            {
                "initial_sss_index": 4,
                "num_cz_gates": 4,
                "00": 0.8,
                "01": 0.0,
                "10": 0.0,
                "11": 0.0,
            },
            {
                "initial_sss_index": 4,
                "num_cz_gates": 5,
                "00": 0.7,
                "01": 0.0,
                "10": 0.0,
                "11": 0.0,
            },
            {
                "initial_sss_index": 4,
                "num_cz_gates": 6,
                "00": 0.6,
                "01": 0.0,
                "10": 0.0,
                "11": 0.0,
            },
        ]
    )

    plot_filename = tmp_path / "example.png"

    results.analyze(plot_filename=plot_filename.as_posix())

    # Calculated by hand
    assert results.cz_fidelity_estimate == pytest.approx(0.81358891968418)
    assert results.cz_fidelity_estimate_std == pytest.approx(0.14459611416597062)

    assert pathlib.Path(tmp_path / "example.png").exists()


def test_results_no_data() -> None:
    results = SSBResults(target="example", experiment=MagicMock(), data=None)
    with pytest.raises(RuntimeError, match=r"No data stored. Cannot perform analysis."):
        results._analyze()

    with pytest.raises(RuntimeError, match=r"No data stored. Cannot plot results."):
        results.plot_results()


def test_results_not_analyzed() -> None:
    results = SSBResults(target="example", experiment=MagicMock(), data=None)
    for attr in ["cz_fidelity_estimate", "cz_fidelity_estimate_std"]:
        with pytest.raises(
            RuntimeError,
            match=re.escape("Value has not yet been estimated. Please run `.analyze()` method."),
        ):
            getattr(results, attr)


def test_dump_and_load(
    tmp_path_factory: pytest.TempPathFactory,
    ssb_experiment: SSB,
) -> None:
    filename = tmp_path_factory.mktemp("tempdir") / "file.json"
    ssb_experiment.to_file(filename)
    exp = SSB.from_file(filename)

    assert exp.samples == ssb_experiment.samples
    assert exp.num_qubits == ssb_experiment.num_qubits
    assert exp.num_circuits == ssb_experiment.num_circuits
    assert exp.cycle_depths == ssb_experiment.cycle_depths
    assert exp.samples == ssb_experiment.samples


def test_repr(ssb_experiment: SSB) -> None:
    assert ssb_experiment.__repr__() == "SSB(num_qubits=2, num_samples=30)"
    results = SSBResults(target="target", experiment=ssb_experiment, job=MagicMock(spec=css.Job))
    assert results.__repr__() == (
        "SSBResults(Results not analyzed, experiment=SSB(num_qubits=2, "
        "num_samples=30), target=target)"
    )

    results._cz_fidelity_estimate = 0.98
    results._cz_fidelity_estimate_std = 0.01
    assert results.__repr__() == (
        "SSBResults(Estimated CZ fidelity: 0.98 +/- 0.01, experiment=SSB(num_qubits=2, "
        "num_samples=30), target=target)"
    )
